/*--------------------------------------------------
ESP-Webserver Dateien im SPIFFS
v0.5 MSoft 18.05.2016
32MBit-Flash: 4M(1M SPIFFS), NodeMCU
--------------------------------------------------*/

#include <ESP8266WiFi.h>
#include <ESP8266WebServer.h>
#include <FS.h>
#include <base64.h>

//#define DEBUG

#define WIFI_TIMEOUT 20                                    // Wartezeit für connect * 500ms 
#define MAX_PIC_SIZE  1 * 1024                             // bis zu dieser Größe wird Base64 encoded

String ssid = "xxxxxxx";                                  // default, wird aus config.txt überschrieben wenn vorhanden
String password = "xxxxxxxx";
String bu_ssid = ""; 
String bu_password = "";
String ap_name = "ESP-Test";
String ap_password = "";
String ap_ip   = "192.168.4.1";

bool ap_mode = false;                                      // true wenn AP aktiv  
 
void configWifi(void);                                     // holt die WLAN-Daten aus config.txt
void WiFiStart(void);

String GetPicture(String filename);                        // codiert ein Bild aus dem SPIFFS in Base64 wenn kleiner als MAX_PIC_SIZE
String getContentType(String filename);                    // setzt Content-Typ nach Dateiendung
void notFound(void);                                       // schickt 404-Meldung 

void handleUnknown(void);                                  // alles, was auf SPIFFS gefunden wurde und keine callback-Funktion hat

void handleUpload(void);                                   // Datei-Upload holt File und schreibt ins SPIFFS

void handleConfig(void);
void handleIndex(void);
void handleForm(void);
void handleEingabe(void);
void handleSchalter(void);

// Create an instance of the webserver on Port 80
ESP8266WebServer webserver(80);

// ###############################################################
// --------------------- setup/loop ------------------------------
// ###############################################################

void setup()
{
  // start serial
  Serial.begin(115200);                           
  Serial.println("");
  Serial.println("RESET");
  
  // Initialize file system.
  if (!SPIFFS.begin())
  {
    Serial.println("SPIFFS nicht initialisiert!");   
    while(1)                                        // ohne SPIFFS geht es sowieso nicht...
    {
      yield();
    }
  }
  Serial.println("SPIFFS ok");

  configWifi();
       
  // inital connect
  WiFi.mode(WIFI_STA);
  WiFiStart();

  // Start the server
  webserver.on("/", handleIndex);
  webserver.on("/index.html", handleIndex);
  webserver.on("/index.htm", handleIndex);

  webserver.on("/config.html", HTTP_GET, handleConfig);
  webserver.on("/config.html", HTTP_POST, handleConfig, handleUpload);

  webserver.on("/form.html", handleForm);
  webserver.on("/eingabe.html", handleEingabe);
  webserver.on("/schalter.html", handleSchalter);
  
  webserver.onNotFound(handleUnknown);
  webserver.begin();

  Serial.println("HTTP server gestartet Port 80");
}

// ----------------------------------------------------------------------

void loop()
{
  if (!ap_mode)
  {
    // check if WLAN is connected
    if (WiFi.status() != WL_CONNECTED)
    {
      WiFiStart();
    }
  }  

  webserver.handleClient();
}

// ###############################################################
// ---------------------- WiFi -----------------------------------
// ###############################################################

void configWifi()
{
#ifdef DEBUG
  Serial.println("\r\n--- configWifi()  ---");
#endif
  
  File configfile = SPIFFS.open("/config.txt","r");         // config.txt für WLAN-Zugriff holen
  if (configfile)
  {

    while(String indaten = configfile.readStringUntil('\n'))
    {
      if (indaten.length() == 0)
      {
        break;
      }  
      if (indaten.startsWith("user"))           // username WLAN-Client
      {
        indaten.replace("user","");
        indaten.trim();
        ssid = indaten;  
      }
      else if (indaten.startsWith("pass"))           // username WLAN-Client
      {
        indaten.replace("pass","");
        indaten.trim();
        password = indaten;
      }
      else if (indaten.startsWith("buser"))           // username WLAN-Client
      {
        indaten.replace("buser","");
        indaten.trim();
        bu_ssid = indaten;  
      }
      else if (indaten.startsWith("bpass"))           // username WLAN-Client
      {
        indaten.replace("bpass","");
        indaten.trim();
        bu_password = indaten;
      }
      else if (indaten.startsWith("ap_name"))           // username WLAN-Client
      {
        indaten.replace("ap_name","");
        indaten.trim();
        ap_name = indaten;
      }
      else if (indaten.startsWith("ap_pass"))           // username WLAN-Client
      {
        indaten.replace("ap_pass","");
        indaten.trim();
        ap_password = indaten;
      }
      else if (indaten.startsWith("ap_ip"))           // username WLAN-Client
      {
        indaten.replace("ap_ip","");
        indaten.trim();
        ap_ip = indaten;
      }
    }  
    configfile.close();
  }
}

// ----------------------------------------------------------------------

void WiFiStart()
{
  char ssid_send[64];
  char password_send[64];

  // Connect to WiFi network
  ssid.toCharArray(ssid_send,ssid.length() + 1);
  password.toCharArray(password_send,password.length() + 1);

  WiFi.begin(ssid_send, password_send);
  Serial.print("\r\nWiFi\r\n");
  Serial.println(ssid_send);
  Serial.println(password_send);
  for (int i=0; i<WIFI_TIMEOUT; i++)
  {
    if (WiFi.status() != WL_CONNECTED)
    {
      Serial.print(".");
      delay(500);
    }
  }  
  if (WiFi.status() == WL_CONNECTED)
  {
    Serial.println("connected");
    IPAddress myAddr = WiFi.localIP();
    Serial.println(myAddr);
  }
  else
  {
    bu_ssid.toCharArray(ssid_send,bu_ssid.length() + 1);
    bu_password.toCharArray(password_send,bu_password.length() + 1);

    WiFi.begin(ssid_send, password_send);
    Serial.print("\r\nWiFi BU\r\n");
    Serial.println(ssid_send);
    Serial.println(password_send);
    for (int i=0; i<WIFI_TIMEOUT; i++)
    {
      if (WiFi.status() != WL_CONNECTED)
      {
        Serial.print(".");
        delay(500);
      }
     if (WiFi.status() == WL_CONNECTED)
      {
        Serial.println("BU connected");
        IPAddress myAddr = WiFi.localIP();
        Serial.println(myAddr);
      }
    }
  }  
  if (WiFi.status() != WL_CONNECTED)
  {  
    WiFi.mode(WIFI_AP);

    ap_name.toCharArray(ssid_send,ap_name.length() + 1);
    ap_password.toCharArray(password_send,ap_password.length() + 1);

    if (ap_ip.length() > 0)
    {
//      WiFi.softAPIP(ap_ip)
    }
    WiFi.softAP(ssid_send, password_send);
    Serial.println("\r\nAP-Mode Start");
    Serial.println(ssid_send);
    Serial.println(password_send);
    IPAddress myAddr = WiFi.softAPIP();
    Serial.println(myAddr);
    ap_mode = true;
  }   
}

// ###############################################################
// ---------------------- Web-Tools ------------------------------
// ###############################################################

void handleUnknown()
{
  String filename = webserver.uri();

#ifdef DEBUG
  Serial.print("URL: ");
  Serial.println(filename);
#endif

  File pageFile = SPIFFS.open(filename, "r"); 
  if (pageFile)
  {
    String contentTyp = getContentType(filename);

#ifdef DEBUG
    Serial.print("Content-Typ: ");
    Serial.println(contentTyp);
#endif

    size_t sent = webserver.streamFile(pageFile, contentTyp);
    pageFile.close(); 
  }
  else
  {
    notFound(); 
  }
}

//---------------------------------------------------------------

void notFound()
{
  String sHTML = F("<html><head><title>404 Not Found</title></head><body><h1>Not Found</h1><p>The requested URL was not found on this webserver.</p></body></html>");
  webserver.send(404, "text/html", sHTML);    
}

//---------------------------------------------------------------

void handleUpload()
{

#ifdef DEBUG
  Serial.println("Datei-Upload"); 
#endif

  static File fsUploadFile;                                         // aktueller Upload

  HTTPUpload& upload = webserver.upload();
  if(upload.status == UPLOAD_FILE_START)
  {
    String dateiname = upload.filename;
    if(!dateiname.startsWith("/"))
    {
      dateiname = "/" + dateiname;
    }
    
#ifdef DEBUG
    Serial.print("handleFileUpload Name: ");
    Serial.println(dateiname);
#endif

    fsUploadFile = SPIFFS.open(dateiname, "w");
    dateiname = String();
  }
  else if(upload.status == UPLOAD_FILE_WRITE)
  {
    if(fsUploadFile)
    {
      fsUploadFile.write(upload.buf, upload.currentSize);

#ifdef DEBUG
      Serial.println(upload.currentSize,DEC);
#endif

    }  
  }
  else if(upload.status == UPLOAD_FILE_END)
  {
    if(fsUploadFile)
    {
      fsUploadFile.close();
    }  

#ifdef DEBUG
    Serial.print("handleFileUpload Size: ");
    Serial.println(upload.totalSize);
#endif

  }
}

// ----------------------------------------------------------------------

String GetPicture(String filename)
{   
  String contentTyp = getContentType(filename);

  File tempFile = SPIFFS.open(filename,"r");     
  if (tempFile)
  {
    if (tempFile.size() < MAX_PIC_SIZE)
    {        
      filename = tempFile.readString(); 
      filename = "data:" + contentTyp + ";base64," + base64::encode((uint8_t *) filename.c_str(), tempFile.size());  
      Serial.println("Test");
      Serial.println(filename);
    }
    tempFile.close();
  }
  return filename;
}

// ----------------------------------------------------------------------

String getContentType(String filename)
{
  if(webserver.hasArg("download"))
    return "application/octet-stream";
  else if(filename.endsWith(".ico"))
    return "image/x-icon";
  else if(filename.endsWith(".htm"))
    return "text/html";
  else if(filename.endsWith(".html"))
    return "text/html";
  else if(filename.endsWith(".css"))
    return "text/css";
  else if(filename.endsWith(".js"))
    return "application/javascript";
  else if(filename.endsWith(".png"))
    return "image/png";
  else if(filename.endsWith(".gif"))
    return "image/gif";
  else if(filename.endsWith(".jpg"))
    return "image/jpeg";
  else if(filename.endsWith(".ico"))
    return "image/x-icon";
  else if(filename.endsWith(".xml"))
    return "text/xml";
  else if(filename.endsWith(".pdf"))
    return "application/x-pdf";
  else if(filename.endsWith(".zip"))
    return "application/x-zip";
  else if(filename.endsWith(".gz"))
    return "application/x-gzip";
  return "text/plain";
}

// ###############################################################
//---------------------- Webseiten -------------------------------
// ###############################################################

void handleIndex()
{
  String filename = webserver.uri();
  if (filename == "/")
  {
    filename = "/index.html";
  }

#ifdef DEBUG
  Serial.print("URL: ");
  Serial.println(filename);
#endif

  File pageFile = SPIFFS.open(filename, "r"); 
  if (pageFile)
  {
    String sHTML = pageFile.readString();
    pageFile.close();  

    Serial.println("Replace...");

    sHTML.replace("~t_bz~", "21,3");
    sHTML.replace("~f_bz~", "48");
    sHTML.replace("~t_ba~", "22,4");
    sHTML.replace("~f_ba~", "52");
    sHTML.replace("~t_wz~", "22,4");
    sHTML.replace("~f_wz~", "49");
    sHTML.replace("~t_ke~", "18,2");
    sHTML.replace("~f_ke~", "51");
    sHTML.replace("~t_sz~", "21,6");
    sHTML.replace("~t_ge~", "-21,8");

    sHTML.replace("~druck~", "1018");

    sHTML.replace("~e_nas~", "32,2");
    sHTML.replace("~e_kuehl~", "0,0");
    sHTML.replace("~e_tv~", "1,8");
    sHTML.replace("~e_bastel~", "7,1");
    sHTML.replace("~e_pc~", "92,4");
    sHTML.replace("~e_uhr~", "3,7");
       
    webserver.send(200, "text/html", sHTML);
  }
  else
  {
    notFound(); 
  }
}

// ----------------------------------------------------------------------

void handleConfig()
{ 
  static File fsUploadFile;                                         // aktueller Upload

  String filename = webserver.uri();

  File pageFile = SPIFFS.open(filename, "r"); 
  if (pageFile)
  {
    String sHTML = pageFile.readString();
    pageFile.close();

    String bilddaten = "/logo.jpg";
    bilddaten = GetPicture(bilddaten);
    sHTML.replace("~logo~",bilddaten);                              // Bild Webseite einfügen
        
    if (webserver.hasArg("reboot"))                           
    {
      Serial.println("\r\n\r\n--- Neustart ---\r\n");
      ESP.restart();
    }
       
    // config.txt speichern
    File configfile;
    if (webserver.hasArg("config"))                            // neue config.txt
    {

#ifdef DEBUG
      Serial.println("\r\n--- config.txt begin ---");
      Serial.print(webserver.arg("config"));  
      Serial.println("--- config.txt end ---\r\n");
#endif

      configfile = SPIFFS.open("/config.txt","w");             // config.txt überschreiben
      if (configfile)
      {
        configfile.print(webserver.arg("config"));             // Daten schreiben    
        configfile.close();   
        sHTML.replace("~config~",webserver.arg("config"));     // und in Webseite einfügen
      }
    }
    else                                                       // keine configdaten zu schreiben
    {
      File configfile = SPIFFS.open("/config.txt","r");        // config.txt holen
      if (configfile)
      {     
        sHTML.replace("~config~",configfile.readString());     // und in Webseite einfügen
        configfile.close();
      }
      else
      {
        sHTML.replace("~config~","");                          // keine config.txt da
      }
    }

    // Datei download
    if (webserver.hasArg("filename"))                          // Download oder Delete
    {
      if (webserver.hasArg("download"))
      {
        Serial.print("Download Filename: ");
        File downfile = SPIFFS.open(webserver.arg("filename"),"r"); 
        String sendname = webserver.arg("filename");
        sendname.replace("/",""); 
        Serial.println(sendname);
        webserver.sendHeader("Content-Disposition"," attachment; filename=\"" + sendname + "\"");
        size_t sent = webserver.streamFile(downfile,"application/octet-stream");
        downfile.close();
      }
      if (webserver.hasArg("delete"))
      {
        Serial.print("Delete Filename: ");
        SPIFFS.remove(webserver.arg("filename"));
      }
      Serial.println(webserver.arg("filename"));
    }
    
#ifdef DEBUG
    Serial.println("\r\n--- Dir-Liste begin ---");
#endif

    // Dateiliste anzeigen
    Dir dir = SPIFFS.openDir("/");
    int anzahl = 0;
    while (dir.next())
    {
      String filename = dir.fileName(); 
      File dirFile = dir.openFile("r");
      size_t filelen =  dirFile.size(); 

#ifdef DEBUG
      Serial.print(filename);
      Serial.print(" - ");
      Serial.println(filelen);
#endif
      String filename_temp = filename;
      filename_temp.replace("/","");
      String dirzeile = "<form action=\"config.html\" method=\"POST\">\r\n<tr><td align=\"center\">";
      dirzeile += "<input type=\"hidden\" name = \"filename\" value=\"" + filename + "\">" + filename_temp;
      dirzeile +=  "</td>\r\n<td align=\"right\">" + String(filelen,DEC);
      dirzeile += "</td><td align=\"center\"><button type=\"submit\" name= \"download\">Download</button></td>\r\n";
      dirzeile += "<td align=\"center\" bgcolor=\"red\"><button type=\"submit\" name=\"delete\">&nbsp;&nbsp;&nbsp;Delete&nbsp;&nbsp;&nbsp;</button></td>";
      dirzeile += "</tr></form>\r\n~dirzeile~"; 
      sHTML.replace("~dirzeile~",dirzeile);
      anzahl++;
    }   
    sHTML.replace("~dirzeile~","");

    FSInfo fs_info;
    SPIFFS.info(fs_info);
    sHTML.replace("~anzahl~",String(anzahl,DEC));
    sHTML.replace("~total~",String(fs_info.totalBytes/1024,DEC));
    sHTML.replace("~used~",String(fs_info.usedBytes/1024,DEC));
    sHTML.replace("~free~",String((fs_info.totalBytes - fs_info.usedBytes)/1024,DEC));

#ifdef DEBUG
  Serial.println("--- Dir-Liste end ---\r\n");
#endif
   
    webserver.send(200, "text/html", sHTML);
  }
  else
  {
    notFound(); 
  }
}

// ----------------------------------------------------------------------
 
void handleSchalter()
{
  String filename = webserver.uri();

  File pageFile = SPIFFS.open(filename, "r"); 
  if (pageFile)
  {
    String sHTML = pageFile.readString();
    pageFile.close();

    if (webserver.hasArg("dose1"))
    {
      if (webserver.arg("dose1") == "1") 
      {
//        tx_send(15);            
        Serial.println("Computer Ein");
      }
      else
      {
//        tx_send(14);            
        Serial.println("Computer Aus");
      }
    }  
    if (webserver.hasArg("dose2"))
    {
      if (webserver.arg("dose2") == "1")
      {
//        tx_send(7);            
        Serial.println("TV Ein");
      }
      else
      {
//        tx_send(6);            
        Serial.println("TV Aus");
      }
        
    }
    if (webserver.hasArg("dose3"))
    {
      if (webserver.arg("dose3") == "1")
      {
//        tx_send(11);            
        Serial.println("Lampe Ein");
      }
      else
      {
//        tx_send(10);            
        Serial.println("Lampe Aus");
      }
    }  
    if (webserver.hasArg("dose4"))
    {
      if (webserver.arg("dose4") == "1") 
      {
//        tx_send(3);            
        Serial.println("LED Ein");
      }
      else
      {
//        tx_send(2);            
        Serial.println("LED Aus");
      }
    }    
    webserver.send(200, "text/html", sHTML);
  }
  else
  {
    notFound(); 
  }
}

//---------------------------------------------------------------

void handleEingabe()
{ 
  String filename = webserver.uri();

  File pageFile = SPIFFS.open(filename, "r"); 
  if (pageFile)
  {
    String sHTML = pageFile.readString();
    pageFile.close();
    if (webserver.hasArg("display"))
    {
      sHTML.replace("~display~",webserver.arg("display"));
    }
    else
    {
      sHTML.replace("~display~","");
    }
    
    if (webserver.hasArg("speed"))
    {
      sHTML.replace("~speed~",webserver.arg("speed"));
    }
    else
    {
      sHTML.replace("~speed~","");
    }

    webserver.send(200, "text/html", sHTML);
  }
  else
  {
    notFound(); 
  }
}

//---------------------------------------------------------------

void handleForm()
{
  String filename = webserver.uri();

  File pageFile = SPIFFS.open(filename, "r"); 
  if (pageFile)
  {
    String sHTML = pageFile.readString();
    pageFile.close();
    
    String name;
    if (webserver.hasArg("vname"))
    {
      name = webserver.arg("vname"); 
    }
    if (webserver.hasArg("zname"))
    {
      name += " " + webserver.arg("zname");  
    }
    name.trim();
    if (name.length() > 0)
    {
      sHTML.replace("~name~","Sie hei&szlig;en " + name + ".<br>");
    }
    else
    {
      sHTML.replace("~name~","");
    }

    String zutaten;
    if (webserver.hasArg("zutat1"))
    {
      zutaten += " und " + webserver.arg("zutat1"); 
    }
    if (webserver.hasArg("zutat2"))
    {
      zutaten += " und " + webserver.arg("zutat2"); 
    }
    if (webserver.hasArg("zutat3"))
    {
      zutaten += " und " + webserver.arg("zutat3"); 
    }
     
    if (zutaten.startsWith(" und "))
    {
      String temp = zutaten.substring(5);
      sHTML.replace("~zutaten~","Sie haben " + temp + " gew&auml;hlt.<br>");
    }
    else
    {
      sHTML.replace("~zutaten~","");
    }
    
    if (webserver.hasArg("Zahlmethode"))
    {
      sHTML.replace("~zahlweise~","Sie wollen mit " + webserver.arg("Zahlmethode") + " bezahlen.");     
    }
    {
      sHTML.replace("~zahlweise~","");
    }
        
    webserver.send(200, "text/html", sHTML);
  }
  else
  {
    notFound(); 
  }
}

//---------------------------------------------------------------


